/************************************************************************************

	File: lcd.c

	Copyright (C) 2012 S.Suwa <www.suwa-koubou.jp>


	Control for I2C low voltage Lcd module 


************************************************************************************/

/** INCLUDES ***********************************************************************/
#include <p30f3012.h>
#include "typedef.h"
#include "pic_232.h"
#include "pic_print.h"
#include "pic_delay.h"
#include "lcd.h"

/** DEFINITION *********************************************************************/
#define	LCD_DEVICE	0b01111100	// 0x7C (0x3E)
#define	MW	0	// I2C master write
#define	MR	1	// I2C master read

//#define	I2C_HANDLING

/*-----------------------------------------------------------------------------------
	lcd initialize
-----------------------------------------------------------------------------------*/

void	lcd_initialize(void)
{
	unsigned char contrast;

	//*** lcd initialize **********************************************
	// set contrast control value
	contrast = 0b111100;

	// power on after 40msec
	delay_ms(40);

	// function 8bit,N=1,5*8dot
	lcd_cmdwrite(cmd_SET_FUNCTION | prm_8BIT | prm_2LINE | prm_SMALL_FONT | prm_NORMAL_INSTRCT);

	// select extension instruction
	lcd_cmdwrite(cmd_SET_FUNCTION | prm_8BIT | prm_2LINE | prm_SMALL_FONT | prm_EXT_INSTRCT);

	// internal OSC frequency adjustment, by default value(0x04)
	lcd_cmdwrite(cmd_OSC_ADJ | 0x04);	

	// set contrast low
	lcd_cmdwrite(cmd_CONTRAST_L | (contrast & 0b1111));

	// set contrast high and icon turned on
	lcd_cmdwrite(cmd_ICON_CONTRAST_H | prm_ICON_ON | ((contrast >> 4) & 0b11));

	// fllower control
	lcd_cmdwrite(cmd_FOLLOWER);

	// wait for LCD device stable
	delay_ms(200);

	// function 8bit,N=1,5*8dot
	lcd_cmdwrite(cmd_SET_FUNCTION | prm_8BIT | prm_2LINE | prm_SMALL_FONT | prm_NORMAL_INSTRCT);

	// Display ON
	lcd_cmdwrite(cmd_DISPLAY | prm_DISPLAY_ON | prm_CURSOR_OFF | prm_BLINK_OFF); // Display On

	// Clear Display
	lcd_cmdwrite(cmd_CLEAR);	// Clear Display
	delay_ms(2);		// wait for clear display

	// Battery ICON on
	lcd_cmdwrite(cmd_SET_FUNCTION | prm_8BIT | prm_2LINE | prm_SMALL_FONT | prm_EXT_INSTRCT);
	lcd_cmdwrite(cmd_SET_ICONADR +0x0d); // set ICON address
	lcd_datawrite(0b11110);	// set ICON bit
	lcd_cmdwrite(cmd_SET_FUNCTION | prm_8BIT | prm_2LINE | prm_SMALL_FONT | prm_NORMAL_INSTRCT);

}

/*-----------------------------------------------------------------------------------
	icon display on/off
-----------------------------------------------------------------------------------*/

void	icon_display(int adr, int bpos, int sw)
{
	lcd_cmdwrite(cmd_SET_FUNCTION | prm_8BIT | prm_2LINE | prm_SMALL_FONT | prm_EXT_INSTRCT);
	lcd_cmdwrite(cmd_SET_ICONADR + adr); // set ICON address
	if(sw == ON){
		lcd_datawrite(bpos);	// set ICON bit
	}else{
		lcd_datawrite(0);		// clear ICON bit
	}
	lcd_cmdwrite(cmd_SET_FUNCTION | prm_8BIT | prm_2LINE | prm_SMALL_FONT | prm_NORMAL_INSTRCT);

}


/*-----------------------------------------------------------------------------------
	telephone icon on/off
-----------------------------------------------------------------------------------*/

void	icon_telephone(int sw)
{
	icon_display(ICON_TELEPHONE_ADR, ICON_TELEPHONE_BPOS, sw);

/**
	lcd_cmdwrite(cmd_SET_FUNCTION | prm_8BIT | prm_2LINE | prm_SMALL_FONT | prm_EXT_INSTRCT);
	lcd_cmdwrite(cmd_SET_ICONADR +2); // set ICON address
	if(sw == ON){
		lcd_datawrite(0b10000);	// set ICON bit
	}else{
		lcd_datawrite(0b00000);	// clear ICON bit
	}
	lcd_cmdwrite(cmd_SET_FUNCTION | prm_8BIT | prm_2LINE | prm_SMALL_FONT | prm_NORMAL_INSTRCT);
**/
}


/*-----------------------------------------------------------------------------------
	bell(speaker sound out) icon on/off
-----------------------------------------------------------------------------------*/

void	icon_bell(int sw)
{
	icon_display(ICON_SPEAKER_ADR, ICON_SPEAKER_BPOS, sw);

/***
	lcd_cmdwrite(cmd_SET_FUNCTION | prm_8BIT | prm_2LINE | prm_SMALL_FONT | prm_EXT_INSTRCT);
	lcd_cmdwrite(cmd_SET_ICONADR +4); // set ICON address
	if(sw == ON){
		lcd_datawrite(0b10000);	// set ICON bit
	}else{
		lcd_datawrite(0b00000);	// clear ICON bit
	}
	lcd_cmdwrite(cmd_SET_FUNCTION | prm_8BIT | prm_2LINE | prm_SMALL_FONT | prm_NORMAL_INSTRCT);
***/
}


/*-----------------------------------------------------------------------------------
	command write
-----------------------------------------------------------------------------------*/

void	lcd_cmdwrite(unsigned char cmd)
{
	i2cOpenStart();
	i2cSend(LCD_DEVICE +MW);
	i2cSend(0b00000000);	// RS=0, select instruction register
	i2cSend(cmd);
	i2cStopClose();
	delay_ms(1);
}


/*-----------------------------------------------------------------------------------
	data write
-----------------------------------------------------------------------------------*/

void	lcd_datawrite(char data)
{
	i2cOpenStart();
	i2cSend(LCD_DEVICE +MW);
	i2cSend(0b01000000);	// RS=1, select data register
	i2cSend((uchar)data);
	i2cStopClose();
	delay_ms(1);
}


/*-----------------------------------------------------------------------------------
	string write
-----------------------------------------------------------------------------------*/

void	lcd_strwrite(char *str)
{
	while(*str){
		lcd_datawrite(*str++);
	}
}


#ifdef	I2C_HANDLING

/*==========================================================================
	I2C Handling module
==========================================================================*/
#define	TIME_OUT	0xffff
static	unsigned short tw;

/*--------------------------------------------------------------------------
	I2C master Open and Start
--------------------------------------------------------------------------*/

void	i2cOpenStart(void)
{
	// enabled I2C
	I2CCONbits.I2CEN = 1;
	//delay_1us();

	// START 
	I2CCONbits.SEN = 1;
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2CCONbits.SEN) break;
	}
}


/*--------------------------------------------------------------------------
	I2C master ReStart
--------------------------------------------------------------------------*/

void	i2cReStart(void)
{
	// RESTART 
	I2CCONbits.RSEN = 1;
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2CCONbits.RSEN) break;
	}
}


/*--------------------------------------------------------------------------
	I2C master Stop and Close
--------------------------------------------------------------------------*/

void	i2cStopClose(void)
{
	// STOP 
	I2CCONbits.PEN = 1;
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2CCONbits.PEN)break;
	}
	//delay_1us();

	// disabled I2C
	I2CCONbits.I2CEN = 0;
}


/*--------------------------------------------------------------------------
	I2C master data send
--------------------------------------------------------------------------*/

int	i2cSend(unsigned char data)
{
	I2CTRN = data;

    if(I2CSTATbits.IWCOL){	// If write collision occurs,return -2
		return -2;
    }

	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2CSTATbits.TRSTAT) break;  // wait until write cycle is complete 
	}

	if (I2CSTATbits.ACKSTAT){ 	// test for ACK condition received
		return -1;	// NACK received
	}

	return 0;	// complete with no error.
}


/*--------------------------------------------------------------------------
	I2C	master data receive
--------------------------------------------------------------------------*/

unsigned char	i2cRec(void)
{
	// enabled Receive
	I2CCONbits.RCEN = 1;

	// wait until receive complete and data stored to I2CRCV register
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2CCONbits.RCEN && I2CSTATbits.RBF) break;
	}

	// clear overflow bit
	I2CSTATbits.I2COV = 0;
    
	return I2CRCV;
}


/*--------------------------------------------------------------------------
	I2C master Ack send
--------------------------------------------------------------------------*/

void	i2cAckSend(void)
{
	// Ack Send
	I2CCONbits.ACKDT = 0;
	I2CCONbits.ACKEN = 1;

	// wait until ack cycle complete
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2CCONbits.ACKEN)break;
	}
}


/*--------------------------------------------------------------------------
	I2C master NotAck send
--------------------------------------------------------------------------*/

void	i2cNAckSend(void)
{
	// NAck Send
	I2CCONbits.ACKDT = 1;
	I2CCONbits.ACKEN = 1;

	// wait until ack cycle complete
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2CCONbits.ACKEN) break;
	}
}
#endif // I2C_HANDLING

/*** end of lcd.c **************************************************************/
