/**************************************************************************
	rtc8564.c

	Real Time Clock module, EPSONTOYOCOM RTC-8564

***************************************************************************/

/** INCLUDES **************************************************************/
#include <p30f3012.h>
#include "pic_232.h"
#include "pic_print.h"
#include "pic_delay.h"
#include "rtc8564.h"

/** DEFINITION ************************************************************/
// rtc device address
#define	RTCDEVADR	0xA2

// register address
#define	RTCCON1		0
#define	RTCCON2		1
#define	RTCSEC		2
#define	RTCMIN		3
#define	RTCHOUR		4
#define	RTCDAY		5
#define	RTCWDAY		6
#define	RTCMONTH	7
#define RTCYEAR		8
#define	RTCALMIN	9
#define	RTCALHOUR	10
#define	RTCALDAY	11
#define	RTCALWDAY	12
#define	RTCCLKOUT	13
#define	RTCTMRCON	14
#define	RTCTIMER	15

// Start/Stop
#define	RTCSTART	1
#define	RTCSTOP		0

// I2C data transfer mode bit
#define	MW			0	// I2C master write mode
#define MR			1	// I2C master read mode

#define	I2C_HANDLING	// use i2c access routines in this file

/*==========================================================================
	INT1 interrupt
	RTC interval interrupt
==========================================================================*/

volatile unsigned char rtc_interrupt;

void __attribute__ ((__interrupt__, no_auto_psv)) _INT1Interrupt(void)
{
	rtc_interrupt = 1;		// set interrupt flag
	IFS1bits.INT1IF = 0;	// clear interrupt flag
}


//#define	RTCC_TEST
#ifdef	RTCC_TEST

/*==========================================================================
	RTCC Test
==========================================================================*/
void	write_real_time_clock(void);
void	read_real_time_clock(void);
void	set_interval_timer(void);
void	clear_interval_timer(void);

void	rtc8564_test(void)
{
	int	c;


	rs_putstr("\r\n**** RTC-8564 Test ************************************\r\n");
	for(;;){
		rs_putstr(" I)nitialize, W)rite time, R)ead time, \r\n");
		rs_putstr(" S)et interval timer inturrupt, C)lear interrupt.\r\n");
		rs_putstr("When Ctrl+C key hits, return to system.\r\n");

		while((c = rs_getwchr()) == -1);

		switch(c){
		case 0x03:
			goto exit;

		case 'I':
		case 'i':
			rtcInit();
			rs_putstr(" ... initialize complete.\r\n");
			break;

		case 'W':
		case 'w':
			write_real_time_clock();	
			break;

		case 'R':
		case 'r':
			read_real_time_clock();	
			break;

		case 'S':
		case 's':
			set_interval_timer();
			break;

		case 'C':
		case 'c':
			clear_interval_timer();
			break;


		}
	}
exit:
	rs_putstr("... return to System.\r\n");

}

/*--------------------------------------------------------------------------
	clear interval timer
--------------------------------------------------------------------------*/

void	clear_interval_timer(void)
{

	rtcClearIntervalTimer();
	rs_putstr("\r\n... timer clear complete.\r\n");

}


/*--------------------------------------------------------------------------
	set interval timer interrupt
--------------------------------------------------------------------------*/

void	set_interval_timer(void)
{
	unsigned short	val;
	int		c;
	unsigned char data[7];
	char	pbuf[10];

	rs_set_echo(ECHO_ON);
	rs_putstr("input interval time value(second) +CR.=>");

	val = 0;
	for(val = 0;;){
		c = rs_getwchr();

		if(c == 0x03){
			rs_putstr("\r\n ... input cancel.\r\n");
			rs_set_echo(ECHO_OFF);
			return;
		}

		if(c == '\r') break;

		if(c >= '0' && c <= '9'){
			val = val *10 + c - '0';
		}
	}
	rs_set_echo(ECHO_OFF);

	if(val > 255){
		rs_putstr("input error, valid value is 1 to 255.\r\n");
		return;
	}

	rtcSetIntervalTimer(val);

	for(;;){
		// wait interrupt
		rtc_interrupt = 0;
		while(rtc_interrupt == 0 && (c = rs_getchr()) != 0x03);
		if(c == 0x03) break;

		rtcReadTime(data);
		rs_putstr("Interrupt! ");
		// display time.
		rs_putstr(int_to_2hex(pbuf, data[2]));	// hour
		rs_putstr(":");
		rs_putstr(int_to_2hex(pbuf, data[1]));	// min
		rs_putstr(":");
		rs_putstr(int_to_2hex(pbuf, data[0]));	// sec
		rs_putstr("\r\n");
	}
	rs_putstr("\r\n... return to Test Menu.\r\n");
}



/*--------------------------------------------------------------------------
	read real time clock value
--------------------------------------------------------------------------*/

void	read_real_time_clock(void)
{
	char pbuf[32];
	unsigned char data[2][7];
	unsigned char prev_sec;
	int		prev, i;

	rs_putstr("Hit Ctrl+C, return to Test Menu.\r\n");
	prev_sec = 0xff;

	for(;;){

		if(rs_getchr() == 0x03) break;

		// first read
		rtcReadTime(&(data[0][0]));
		prev = 1;
		for(;;){
			// next read
			rtcReadTime(&(data[prev][0]));
			for(i = 0; i < 7; i++){
				if(data[0][i] != data[1][i]) break;
			}
			if(i >= 7) break;	// match break
			prev ^= 1;	// read retry
		}	

		// changed second value?
		if(prev_sec != data[0][0]){
			prev_sec = data[0][0];
			// display time.
			rs_putstr("20");
			rs_putstr(int_to_2hex(pbuf, data[0][6]));	// year
			rs_putstr("/");
			rs_putstr(int_to_2hex(pbuf, data[0][5]));	// month
			rs_putstr("/");
			rs_putstr(int_to_2hex(pbuf, data[0][3]));	// day
			rs_putstr("(");
			rs_putstr(int_to_2hex(pbuf, data[0][4]));	// wday
			rs_putstr(") ");
			rs_putstr(int_to_2hex(pbuf, data[0][2]));	// hour
			rs_putstr(":");
			rs_putstr(int_to_2hex(pbuf, data[0][1]));	// min
			rs_putstr(":");
			rs_putstr(int_to_2hex(pbuf, data[0][0]));	// sec
			rs_putstr("\r");
		}
	}
	rs_putstr("\r\n... return to Test Menu.\r\n");

}



/*--------------------------------------------------------------------------
	write real time clock value
--------------------------------------------------------------------------*/

void	write_real_time_clock(void)
{
	unsigned char data[7];
	char	buf[20];
	int		c, i;


	rs_set_echo(ECHO_ON);
	rs_putstr("Input time value +CR (ex. 2011071606235959)=>");

	for(i = 0;;){
		while((c = rs_getwchr()) == -1);

		if(c == 0x03){
			rs_putstr("\r\n ... input cancel.\r\n");
			rs_set_echo(ECHO_OFF);
			return;
		}

		if(c == '\r'){
			if(i != 16){
				rs_putstr("\r\n ... input error.\r\n");
				rs_set_echo(ECHO_OFF);
				return;
			}
			break;
		}

		if(i < 16 && c >= '0' && c <= '9'){
			buf[i] = c - '0';
			i++;
		}
	}

	rs_set_echo(ECHO_OFF);
	data[0] = buf[14] << 4 | buf[15];	// second
	data[1] = buf[12] << 4 | buf[13];	// minute
	data[2] = buf[10] << 4 | buf[11];	// hour
	data[4] = buf[8] << 4 | buf[9];		// wday
	data[3] = buf[6] << 4 | buf[7];		// day
	data[5] = buf[4] << 4 | buf[5];		// month
	data[6] = buf[2] << 4 | buf[3];		// year
	rtcWriteTime(data);

	rs_putstr("\r\n... time set complete.\r\n");
}


#endif // RTCC_TEST

/*==========================================================================
	Real Time Clock module, EPSONTOYOCOM RTC-8564
==========================================================================*/

/*--------------------------------------------------------------------------
	RTC initialize
--------------------------------------------------------------------------*/

void	rtcInit(void)
{
	delay_ms(100);

	i2cOpenStart();
	i2cSend(RTCDEVADR + MW);
	i2cSend(RTCCON1);
	i2cSend(0x20);	// stop
	i2cSend(0x00);	// clear flag, disable interrupt
	i2cSend(0x59);	// 59sec
	i2cSend(0x59);	// 59min
	i2cSend(0x23);	// 23hour
	i2cSend(0x31);	// 31day
	i2cSend(0x05);	// 2010/12/31(friday)
	i2cSend(0x12);	// 12month
	i2cSend(0x10);	// 2010year
	i2cSend(0x00);	// disable alarm min
	i2cSend(0x00);	// disable alarm hour
	i2cSend(0x01);	// disable alarm day
	i2cSend(0x00);	// disable alarm wday
	i2cSend(0x00);	// CLKOUT off
	i2cSend(0x00);	// disable timer interrupt
	i2cSend(0xFF);	// timer count value is max
	i2cStopClose();

	delay_us(5);

	rtcCont(RTCSTART);	// start clock count

}


/*--------------------------------------------------------------------------
	RTC Start/Stop
--------------------------------------------------------------------------*/

void	rtcCont(int sw)
{
	i2cOpenStart();
	i2cSend(RTCDEVADR + MW);
	i2cSend(RTCCON1);
	i2cSend(sw == RTCSTART ? 0x0: 0x20);
	i2cStopClose();
	delay_us(5);
}


/*--------------------------------------------------------------------------
	RTC Write Time
	tm[0] = sec, tm[1] = min, tm[2] = hour,
	tm[3] = day, tm[4] = wday, tm[5] = month, tm[6] = year;	
--------------------------------------------------------------------------*/

void	rtcWriteTime(unsigned char tm[])
{
	short	i;

	// stop clock
	rtcCont(RTCSTOP);

	// set Date & Time
	i2cOpenStart();
	i2cSend(RTCDEVADR + MW);
	i2cSend(RTCSEC);
	for(i = 0; i < 7; i++){
		i2cSend(tm[i]);
	}
	i2cStopClose();
	delay_us(5);

	// start clock
	rtcCont(RTCSTART);
}	


/*--------------------------------------------------------------------------
	RTC Write Alarm Time
	tm[0] = min, tm[0] = hour,
	tm[2] = day, tm[3] = wday
--------------------------------------------------------------------------*/

void	rtcWriteAlarmTime(unsigned char tm[])
{
	short	i;

	//rtcCont(RTCSTOP);

	// set Alarm time
	i2cOpenStart();
	i2cSend(RTCDEVADR + MW);
	i2cSend(RTCALMIN);
	for(i = 0; i < 4; i++){
		i2cSend(tm[i]);
	}
	i2cStopClose();
	delay_us(5);

	//rtcCont(RTCSTART);

}


/*--------------------------------------------------------------------------
	RTC Read Time
	tm[0] = sec, tm[1] = min, tm[2] = hour,
	tm[3] = day, tm[4] = wday, tm[5] = month, tm[6] = year;	
--------------------------------------------------------------------------*/

void	rtcReadTime(unsigned char tm[])
{
	short	i;

	i2cOpenStart();
	i2cSend(RTCDEVADR + MW);
	i2cSend(RTCSEC);
	i2cReStart();
	i2cSend(RTCDEVADR + MR);
	for(i = 0; i < 7; i++){
		tm[i] = i2cRec();
		if(i < 6){
			i2cAckSend();
		}else{
			i2cNAckSend();
		}
	}
	i2cStopClose();
	//delay_us(5);

	// masked register invalid bits
	tm[0] &= 0x7f;	// second
	tm[1] &= 0x7f;	// minute
	tm[2] &= 0x3f;	// hour
	tm[3] &= 0x3f;	// day
	tm[4] &= 0x07;	// wday
	tm[5] &= 0x1f;	// month
}	


void	rtcReadNowTime(unsigned char tm[])
{
	unsigned char	data[2][7];
	int	i, prev;

	// first read
	rtcReadTime(&(data[0][0]));
	prev = 1;
	for(;;){
		// next read
		rtcReadTime(&(data[prev][0]));
		for(i = 0; i < 7; i++){
			if(data[0][i] != data[1][i]) break;
		}
		if(i >= 7) break;	// match break
		prev ^= 1;	// read retry
	}	

	for(i = 0; i < 7; i++){
		tm[i] = data[0][i];
	}
}


/*--------------------------------------------------------------------------
	RTC Read Alarm Time
	tm[0] = min, tm[0] = hour,
	tm[2] = day, tm[3] = wday
--------------------------------------------------------------------------*/

void	rtcReadAlarmTime(unsigned char tm[])
{
	short	i;

	i2cOpenStart();
	i2cSend(RTCDEVADR + MW);
	i2cSend(RTCALMIN);
	i2cReStart();
	i2cSend(RTCDEVADR + MR);
	for(i = 0; i < 4; i++){
		tm[i] = i2cRec();
		if(i < 3){
			i2cAckSend();
		}else{
			i2cNAckSend();
		}
	}
	i2cStopClose();
	//delay_us(5);

	// masked register invalid bits
	tm[0] &= 0x7f;	// alarm min
	tm[1] &= 0x3f;	// alarm hour
	tm[2] &= 0x3f;	// alarm day
	tm[3] &= 0x07;	// alarm wday
}	


/*--------------------------------------------------------------------------
	RTC Set Interval Timer, unit sec.
--------------------------------------------------------------------------*/

void	rtcSetIntervalTimer(unsigned char val)
{
	unsigned char	control2;

	// timer stop, and select source clock
	i2cOpenStart();
	i2cSend(RTCDEVADR + MW);
	i2cSend(RTCTMRCON);
	i2cSend(0x02);	// disable interval timer(TE=0), source clock is 1Hz(TD1,0=10)
	i2cStopClose();
	delay_us(5);

	// read control2
	i2cOpenStart();
	i2cSend(RTCDEVADR + MW);
	i2cSend(RTCCON2);
	i2cReStart();
	i2cSend(RTCDEVADR + MR);
	control2 = i2cRec();
	i2cNAckSend();
	i2cStopClose();
	delay_us(5);

	// clear TF, set TIE
	control2 &= 0b00001010;	// clear TF,TIE
	control2 |= 0b00010001;	// set TP(repeated), TIE(Timer Interrupt Enable)

	// set interrupt type
	i2cOpenStart();
	i2cSend(RTCDEVADR + MW);
	i2cSend(RTCCON2);
	i2cSend(control2);
	i2cStopClose();
	delay_us(5);

	// set interval value
	i2cOpenStart();
	i2cSend(RTCDEVADR + MW);
	i2cSend(RTCTIMER);
	i2cSend(val);
	i2cStopClose();
	delay_us(5);

	// timer start, and select source clock
	i2cOpenStart();
	i2cSend(RTCDEVADR + MW);
	i2cSend(RTCTMRCON);
	i2cSend(0x82);	// enable interval timer(TE=1), source clock is 1Hz(TD1,0=10)
	i2cStopClose();
	delay_us(5);
}


/*--------------------------------------------------------------------------
	RTC Clear Interval Timer.(disable interrupt timer)
--------------------------------------------------------------------------*/

void	rtcClearIntervalTimer(void)
{
	// timer stop
	i2cOpenStart();
	i2cSend(RTCDEVADR + MW);
	i2cSend(RTCTMRCON);
	i2cSend(0x0);
	i2cStopClose();
	delay_us(5);
}


/*--------------------------------------------------------------------------
	RTC Set Alarm
--------------------------------------------------------------------------*/

void	rtcSetAlarm(void)
{
	unsigned char	control2;

	// stop RTCC
	//rtcCont(RTCSTOP);

	// read control2
	i2cOpenStart();
	i2cSend(RTCDEVADR + MW);
	i2cSend(RTCCON2);
	i2cReStart();
	i2cSend(RTCDEVADR + MR);
	control2 = i2cRec();
	i2cNAckSend();
	i2cStopClose();
	delay_us(5);

	// clear AF, set AIE
	control2 &= 0b00010101;	// clear AF,AIE
	control2 |= 0b00000010;	// set AIE (Alarm Interrupt Enable)

	// write control2
	i2cOpenStart();
	i2cSend(RTCDEVADR + MW);
	i2cSend(RTCCON2);
	i2cSend(control2);
	i2cStopClose();
	delay_us(5);

	// start RTCC
	//rtcCont(RTCSTART);

}


/*--------------------------------------------------------------------------
	RTC Clear Alarm
--------------------------------------------------------------------------*/

void	rtcClearAlarm(void)
{
	unsigned char	control2;

	// stop RTCC
	//rtcCont(RTCSTOP);

	// read control2
	i2cOpenStart();
	i2cSend(RTCDEVADR + MW);
	i2cSend(RTCCON2);
	i2cReStart();
	i2cSend(RTCDEVADR + MR);
	control2 = i2cRec();
	i2cNAckSend();
	i2cStopClose();
	delay_us(5);

	// clear AF, AIE
	control2 &= 0b00010101;	// clear AF,AIE

	// write control2
	i2cOpenStart();
	i2cSend(RTCDEVADR + MW);
	i2cSend(RTCCON2);
	i2cSend(control2);
	i2cStopClose();
	delay_us(5);

	// start RTCC
	//rtcCont(RTCSTART);

}


/*--------------------------------------------------------------------------
	RTC Get Interrupt Flag
--------------------------------------------------------------------------*/

unsigned char	rtcGetIF(void)
{
	unsigned char	control2;

//rs_putstr("enter rtcGetIF()\r\n");

	// read control2
	i2cOpenStart();
	i2cSend(RTCDEVADR + MW);
	i2cSend(RTCCON2);
	i2cReStart();
	i2cSend(RTCDEVADR + MR);
	control2 = i2cRec();
	i2cNAckSend();
	i2cStopClose();
	delay_us(5);

//myprintf("control2=%02x\r\n", control2);
//rs_putstr("leave rtcGetIF()\r\n");

	return control2 & 0x1f;
}

/*--------------------------------------------------------------------------
	RTC Clear Interrupt Flag
--------------------------------------------------------------------------*/

void	rtcClearIF(void)
{
	unsigned char	control2;
//rs_putstr("enter rtcClearIF()\r\n");

	// stop RTCC
	//rtcCont(RTCSTOP);

	// read control2
	i2cOpenStart();
	i2cSend(RTCDEVADR + MW);
	i2cSend(RTCCON2);
	i2cReStart();
	i2cSend(RTCDEVADR + MR);
	control2 = i2cRec();
	i2cNAckSend();
	i2cStopClose();
	delay_us(5);

//myprintf("control2=%02x\r\n", control2);
	control2 &= 0x13;
//myprintf("control2=%02x\r\n", control2);

	// write control2
	i2cOpenStart();
	i2cSend(RTCDEVADR + MW);
	i2cSend(RTCCON2);
	i2cSend(control2);
	i2cStopClose();
	delay_us(5);

	// start RTCC
	//rtcCont(RTCSTART);

//rs_putstr("leave rtcClearIF()\r\n");
}


#ifdef	I2C_HANDLING

/*==========================================================================
	I2C Handling module
==========================================================================*/
#define	TIME_OUT	0xffff
static	unsigned short tw;

/*--------------------------------------------------------------------------
	I2C master Open and Start
--------------------------------------------------------------------------*/

void	i2cOpenStart(void)
{
	// enabled I2C
	I2CCONbits.I2CEN = 1;
	delay_us(5);

	// START 
	I2CCONbits.SEN = 1;
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2CCONbits.SEN) break;
	}
}


/*--------------------------------------------------------------------------
	I2C master ReStart
--------------------------------------------------------------------------*/

void	i2cReStart(void)
{
	// RESTART 
	I2CCONbits.RSEN = 1;
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2CCONbits.RSEN) break;
	}
}


/*--------------------------------------------------------------------------
	I2C master Stop and Close
--------------------------------------------------------------------------*/

void	i2cStopClose(void)
{
	// STOP 
	I2CCONbits.PEN = 1;
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2CCONbits.PEN)break;
	}
	delay_us(5);

	// disabled I2C
	I2CCONbits.I2CEN = 0;
}


/*--------------------------------------------------------------------------
	I2C master data send
--------------------------------------------------------------------------*/

int	i2cSend(unsigned char data)
{
	I2CTRN = data;

    if(I2CSTATbits.IWCOL){	// If write collision occurs,return -2
		return -2;
    }

	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2CSTATbits.TRSTAT) break;  // wait until write cycle is complete 
	}

	if (I2CSTATbits.ACKSTAT){ 	// test for ACK condition received
		return -1;	// NACK received
	}

	return 0;	// complete with no error.
}


/*--------------------------------------------------------------------------
	I2C	master data receive
--------------------------------------------------------------------------*/

unsigned char	i2cRec(void)
{
	// enabled Receive
	I2CCONbits.RCEN = 1;

	// wait until receive complete and data stored to I2CRCV register
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2CCONbits.RCEN && I2CSTATbits.RBF) break;
	}

	// clear overflow bit
	I2CSTATbits.I2COV = 0;
    
	return I2CRCV;
}


/*--------------------------------------------------------------------------
	I2C master Ack send
--------------------------------------------------------------------------*/

void	i2cAckSend(void)
{
	// Ack Send
	I2CCONbits.ACKDT = 0;
	I2CCONbits.ACKEN = 1;

	// wait until ack cycle complete
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2CCONbits.ACKEN)break;
	}
}


/*--------------------------------------------------------------------------
	I2C master NotAck send
--------------------------------------------------------------------------*/

void	i2cNAckSend(void)
{
	// NAck Send
	I2CCONbits.ACKDT = 1;
	I2CCONbits.ACKEN = 1;

	// wait until ack cycle complete
	for(tw = TIME_OUT; tw > 0; tw--){
		if(!I2CCONbits.ACKEN) break;
	}
}

#endif	// I2C_HANDLING

/*** end of rtc8564.c *****************************************************/
